@@ -1,6 +1,9 @@
 import asyncio
+import errno
 import sys
+import traceback
 import warnings
+import weakref
 
 import psycopg2
 from psycopg2.extensions import (
@@ -8,13 +11,14 @@ from psycopg2.extensions import (
 from psycopg2 import extras
 
 from .cursor import Cursor
+from .utils import _ContextManager, PY_35, create_future
 
 
 __all__ = ('connect',)
 
 
 TIMEOUT = 60.0
-PY_34 = sys.version_info >= (3, 4)
+PY_341 = sys.version_info >= (3, 4, 1)
 
 
 @asyncio.coroutine
@@ -35,9 +39,8 @@ def _enable_hstore(conn):
     return tuple(rv0), tuple(rv1)
 
 
-@asyncio.coroutine
-def connect(dsn=None, *, timeout=TIMEOUT, loop=None,
-            enable_json=True, enable_hstore=True, echo=False, **kwargs):
+def connect(dsn=None, *, timeout=TIMEOUT, loop=None, enable_json=True,
+            enable_hstore=True, enable_uuid=True, echo=False, **kwargs):
     """A factory for connecting to PostgreSQL.
 
     The coroutine accepts all parameters that psycopg2.connect() does
@@ -46,10 +49,19 @@ def connect(dsn=None, *, timeout=TIMEOUT, loop=None,
     Returns instantiated Connection object.
 
     """
+    coro = _connect(dsn=dsn, timeout=timeout, loop=loop,
+                    enable_json=enable_json, enable_hstore=enable_hstore,
+                    enable_uuid=enable_uuid, echo=echo, **kwargs)
+    return _ContextManager(coro)
+
+
+@asyncio.coroutine
+def _connect(dsn=None, *, timeout=TIMEOUT, loop=None, enable_json=True,
+             enable_hstore=True, enable_uuid=True, echo=False, **kwargs):
     if loop is None:
         loop = asyncio.get_event_loop()
 
-    waiter = asyncio.Future(loop=loop)
+    waiter = create_future(loop)
     conn = Connection(dsn, loop, timeout, waiter, bool(echo), **kwargs)
     try:
         yield from conn._poll(waiter, timeout)
@@ -58,6 +70,8 @@ def connect(dsn=None, *, timeout=TIMEOUT, loop=None,
         raise
     if enable_json:
         extras.register_default_json(conn._conn)
+    if enable_uuid:
+        extras.register_uuid(conn_or_curs=conn._conn)
     if enable_hstore:
         oids = yield from _enable_hstore(conn)
         if oids is not None:
@@ -74,59 +88,64 @@ class Connection:
 
     """
 
+    _source_traceback = None
+
     def __init__(self, dsn, loop, timeout, waiter, echo, **kwargs):
         self._loop = loop
         self._conn = psycopg2.connect(dsn, async=True, **kwargs)
         self._dsn = self._conn.dsn
-        assert self._conn.isexecuting(), "Is conn async at all???"
+        assert self._conn.isexecuting(), "Is conn an async at all???"
         self._fileno = self._conn.fileno()
         self._timeout = timeout
         self._waiter = waiter
-        self._reading = False
         self._writing = False
         self._echo = echo
-        self._ready()
+        self._notifies = asyncio.Queue(loop=loop)
+        self._weakref = weakref.ref(self)
+        self._loop.add_reader(self._fileno, self._ready, self._weakref)
+        if loop.get_debug():
+            self._source_traceback = traceback.extract_stack(sys._getframe(1))
 
-    def _ready(self):
-        if self._waiter is None:
-            self._fatal_error("Fatal error on aiopg connection: "
-                              "bad state in _ready callback")
+    @staticmethod
+    def _ready(weak_self):
+        self = weak_self()
+        if self is None:
             return
 
+        waiter = self._waiter
+
         try:
             state = self._conn.poll()
+            while self._conn.notifies:
+                notify = self._conn.notifies.pop(0)
+                self._notifies.put_nowait(notify)
         except (psycopg2.Warning, psycopg2.Error) as exc:
-            if self._reading:
-                self._loop.remove_reader(self._fileno)
-                self._reading = False
-            if self._writing:
-                self._loop.remove_writer(self._fileno)
-                self._writing = False
-            if not self._waiter.cancelled():
-                self._waiter.set_exception(exc)
+            try:
+                if self._writing:
+                    self._writing = False
+                    self._loop.remove_writer(self._fileno)
+            except OSError as exc2:
+                if exc2.errno != errno.EBADF:
+                    # EBADF is ok for closed file descriptor
+                    # chain exception otherwise
+                    exc2.__cause__ = exc
+                    exc = exc2
+            if waiter is not None and not waiter.done():
+                waiter.set_exception(exc)
         else:
             if state == POLL_OK:
-                if self._reading:
-                    self._loop.remove_reader(self._fileno)
-                    self._reading = False
                 if self._writing:
                     self._loop.remove_writer(self._fileno)
                     self._writing = False
-                if not self._waiter.cancelled():
-                    self._waiter.set_result(None)
+                if waiter is not None and not waiter.done():
+                    waiter.set_result(None)
             elif state == POLL_READ:
-                if not self._reading:
-                    self._loop.add_reader(self._fileno, self._ready)
-                    self._reading = True
                 if self._writing:
                     self._loop.remove_writer(self._fileno)
                     self._writing = False
             elif state == POLL_WRITE:
-                if self._reading:
-                    self._loop.remove_reader(self._fileno)
-                    self._reading = False
                 if not self._writing:
-                    self._loop.add_writer(self._fileno, self._ready)
+                    self._loop.add_writer(self._fileno, self._ready, weak_self)
                     self._writing = True
             elif state == POLL_ERROR:
                 self._fatal_error("Fatal error on aiopg connection: "
@@ -151,22 +170,36 @@ class Connection:
         if self._waiter is not None:
             raise RuntimeError('%s() called while another coroutine is '
                                'already waiting for incoming data' % func_name)
-        self._waiter = asyncio.Future(loop=self._loop)
+        self._waiter = create_future(self._loop)
         return self._waiter
 
     @asyncio.coroutine
     def _poll(self, waiter, timeout):
         assert waiter is self._waiter, (waiter, self._waiter)
-        self._ready()
+        self._ready(self._weakref)
+
+        @asyncio.coroutine
+        def cancel():
+            if not self._isexecuting():
+                return
+            self._waiter = create_future(self._loop)
+            self._conn.cancel()
+            try:
+                yield from self._waiter
+            except psycopg2.extensions.QueryCanceledError:
+                pass
+
         try:
             yield from asyncio.wait_for(self._waiter, timeout, loop=self._loop)
+        except (asyncio.CancelledError, asyncio.TimeoutError) as exc:
+            yield from asyncio.shield(cancel(), loop=self._loop)
+            raise exc
         finally:
             self._waiter = None
 
     def _isexecuting(self):
         return self._conn.isexecuting()
 
-    @asyncio.coroutine
     def cursor(self, name=None, cursor_factory=None,
                scrollable=None, withhold=False, timeout=None):
         """A coroutine that returns a new cursor object using the connection.
@@ -179,18 +212,26 @@ class Connection:
         psycopg in asynchronous mode.
 
         """
-        if timeout is None:
-            timeout = self._timeout
-
-        impl = yield from self._cursor(name=name,
-                                       cursor_factory=cursor_factory,
-                                       scrollable=scrollable,
-                                       withhold=withhold)
-        return Cursor(self, impl, timeout, self._echo)
+        coro = self._cursor(name=name, cursor_factory=cursor_factory,
+                            scrollable=scrollable, withhold=withhold,
+                            timeout=timeout)
+        return _ContextManager(coro)
 
     @asyncio.coroutine
     def _cursor(self, name=None, cursor_factory=None,
-                scrollable=None, withhold=False):
+                scrollable=None, withhold=False, timeout=None):
+        if timeout is None:
+            timeout = self._timeout
+
+        impl = yield from self._cursor_impl(name=name,
+                                            cursor_factory=cursor_factory,
+                                            scrollable=scrollable,
+                                            withhold=withhold)
+        return Cursor(self, impl, timeout, self._echo)
+
+    @asyncio.coroutine
+    def _cursor_impl(self, name=None, cursor_factory=None,
+                     scrollable=None, withhold=False):
         if cursor_factory is None:
             impl = self._conn.cursor(name=name,
                                      scrollable=scrollable, withhold=withhold)
@@ -203,17 +244,15 @@ class Connection:
         """Remove the connection from the event_loop and close it."""
         # N.B. If connection contains uncommitted transaction the
         # transaction will be discarded
-        if self._reading:
-            self._loop.remove_reader(self._fileno)
-            self._reading = False
+        self._loop.remove_reader(self._fileno)
         if self._writing:
-            self._loop.remove_writer(self._fileno)
             self._writing = False
+            self._loop.remove_writer(self._fileno)
         self._conn.close()
         if self._waiter is not None and not self._waiter.done():
             self._waiter.set_exception(
                 psycopg2.OperationalError("Connection closed"))
-        ret = asyncio.Future(loop=self._loop)
+        ret = create_future(self._loop)
         ret.set_result(None)
         return ret
 
@@ -276,11 +315,24 @@ class Connection:
     @asyncio.coroutine
     def cancel(self, timeout=None):
         """Cancel the current database operation."""
-        waiter = self._create_waiter('cancel')
-        self._conn.cancel()
-        if timeout is None:
-            timeout = self._timeout
-        yield from self._poll(waiter, timeout)
+        if timeout is not None:
+            warnings.warn('timeout parameter is deprecated and never used',
+                          DeprecationWarning)
+        if self._waiter is not None:
+            self._waiter.cancel()
+        if not self._isexecuting():
+            return
+
+        @asyncio.coroutine
+        def cancel():
+            self._waiter = create_future(self._loop)
+            self._conn.cancel()
+            try:
+                yield from self._waiter
+            except psycopg2.extensions.QueryCanceledError:
+                pass
+
+        yield from asyncio.shield(cancel(), loop=self._loop)
 
     @asyncio.coroutine
     def reset(self):
@@ -395,9 +447,33 @@ class Connection:
         """Return echo mode status."""
         return self._echo
 
-    if PY_34:  # pragma: no branch
+    if PY_341:  # pragma: no branch
         def __del__(self):
-            if not self._conn.closed:
+            try:
+                _conn = self._conn
+            except AttributeError:
+                return
+            if _conn is not None and not _conn.closed:
+                self.close()
                 warnings.warn("Unclosed connection {!r}".format(self),
                               ResourceWarning)
-                self.close()
+
+                context = {'connection': self,
+                           'message': 'Unclosed connection'}
+                if self._source_traceback is not None:
+                    context['source_traceback'] = self._source_traceback
+                self._loop.call_exception_handler(context)
+
+    @property
+    def notifies(self):
+        """Return notification queue."""
+        return self._notifies
+
+    if PY_35:  # pragma: no branch
+        @asyncio.coroutine
+        def __aenter__(self):
+            return self
+
+        @asyncio.coroutine
+        def __aexit__(self, exc_type, exc_val, exc_tb):
+            self.close()
