@@ -4,6 +4,7 @@ import warnings
 import psycopg2
 
 from .log import logger
+from .utils import PY_35
 
 
 class Cursor:
@@ -108,8 +109,11 @@ class Cursor:
         except:
             self._conn._waiter = None
             raise
-        else:
+        try:
             yield from self._conn._poll(waiter, timeout)
+        except asyncio.TimeoutError:
+            self._impl.close()
+            raise
 
     @asyncio.coroutine
     def executemany(self, operation, seq_of_parameters):
@@ -372,3 +376,26 @@ class Cursor:
                 raise StopIteration
             else:
                 yield row
+
+    if PY_35:  # pragma: no branch
+
+        @asyncio.coroutine
+        def __aiter__(self):
+            return self
+
+        @asyncio.coroutine
+        def __anext__(self):
+            ret = yield from self.fetchone()
+            if ret is not None:
+                return ret
+            else:
+                raise StopAsyncIteration  # noqa
+
+        @asyncio.coroutine
+        def __aenter__(self):
+            return self
+
+        @asyncio.coroutine
+        def __aexit__(self, exc_type, exc_val, exc_tb):
+            self.close()
+            return
