@@ -8,23 +8,38 @@ from psycopg2.extensions import TRANSACTION_STATUS_IDLE
 
 from .connection import connect, TIMEOUT
 from .log import logger
+from .utils import (PY_35, _PoolContextManager, _PoolConnectionContextManager,
+                    _PoolCursorContextManager, _PoolAcquireContextManager,
+                    ensure_future, create_future)
 
 
-PY_34 = sys.version_info >= (3, 4)
+PY_341 = sys.version_info >= (3, 4, 1)
+
+
+def create_pool(dsn=None, *, minsize=1, maxsize=10,
+                loop=None, timeout=TIMEOUT,
+                enable_json=True, enable_hstore=True, enable_uuid=True,
+                echo=False,
+                **kwargs):
+    coro = _create_pool(dsn=dsn, minsize=minsize, maxsize=maxsize, loop=loop,
+                        timeout=timeout, enable_json=enable_json,
+                        enable_hstore=enable_hstore, enable_uuid=enable_uuid,
+                        echo=echo, **kwargs)
+    return _PoolContextManager(coro)
 
 
 @asyncio.coroutine
-def create_pool(dsn=None, *, minsize=10, maxsize=10,
-                loop=None, timeout=TIMEOUT,
-                enable_json=True, enable_hstore=True,
-                echo=False,
-                **kwargs):
+def _create_pool(dsn=None, *, minsize=1, maxsize=10,
+                 loop=None, timeout=TIMEOUT,
+                 enable_json=True, enable_hstore=True, enable_uuid=True,
+                 echo=False,
+                 **kwargs):
     if loop is None:
         loop = asyncio.get_event_loop()
 
     pool = Pool(dsn, minsize, maxsize, loop, timeout,
                 enable_json=enable_json, enable_hstore=enable_hstore,
-                echo=echo,
+                enable_uuid=enable_uuid, echo=echo,
                 **kwargs)
     if minsize > 0:
         with (yield from pool._cond):
@@ -36,10 +51,10 @@ class Pool(asyncio.AbstractServer):
     """Connection pool"""
 
     def __init__(self, dsn, minsize, maxsize, loop, timeout, *,
-                 enable_json, enable_hstore, echo, **kwargs):
+                 enable_json, enable_hstore, enable_uuid, echo, **kwargs):
         if minsize < 0:
             raise ValueError("minsize should be zero or greater")
-        if maxsize < minsize:
+        if maxsize < minsize and maxsize != 0:
             raise ValueError("maxsize should be not less than minsize")
         self._dsn = dsn
         self._minsize = minsize
@@ -47,10 +62,11 @@ class Pool(asyncio.AbstractServer):
         self._timeout = timeout
         self._enable_json = enable_json
         self._enable_hstore = enable_hstore
+        self._enable_uuid = enable_uuid
         self._echo = echo
         self._conn_kwargs = kwargs
         self._acquiring = 0
-        self._free = collections.deque(maxlen=maxsize)
+        self._free = collections.deque(maxlen=maxsize or None)
         self._cond = asyncio.Condition(loop=loop)
         self._used = set()
         self._terminated = set()
@@ -90,6 +106,10 @@ class Pool(asyncio.AbstractServer):
                 yield from conn.close()
             self._cond.notify()
 
+    @property
+    def closed(self):
+        return self._closed
+
     def close(self):
         """Close pool.
 
@@ -134,9 +154,13 @@ class Pool(asyncio.AbstractServer):
 
         self._closed = True
 
-    @asyncio.coroutine
     def acquire(self):
         """Acquire free connection from the pool."""
+        coro = self._acquire()
+        return _PoolAcquireContextManager(coro, self)
+
+    @asyncio.coroutine
+    def _acquire(self):
         if self._closing:
             raise RuntimeError("Cannot acquire connection after closing pool")
         with (yield from self._cond):
@@ -153,6 +177,16 @@ class Pool(asyncio.AbstractServer):
 
     @asyncio.coroutine
     def _fill_free_pool(self, override_min):
+        # iterate over free connections and remove timeouted ones
+        n, free = 0, len(self._free)
+        while n < free:
+            conn = self._free[-1]
+            if conn.closed:
+                self._free.pop()
+            else:
+                self._free.rotate()
+            n += 1
+
         while self.size < self.minsize:
             self._acquiring += 1
             try:
@@ -160,6 +194,7 @@ class Pool(asyncio.AbstractServer):
                     self._dsn, loop=self._loop, timeout=self._timeout,
                     enable_json=self._enable_json,
                     enable_hstore=self._enable_hstore,
+                    enable_uuid=self._enable_uuid,
                     echo=self._echo,
                     **self._conn_kwargs)
                 # raise exception if pool is closing
@@ -177,6 +212,7 @@ class Pool(asyncio.AbstractServer):
                     self._dsn, loop=self._loop, timeout=self._timeout,
                     enable_json=self._enable_json,
                     enable_hstore=self._enable_hstore,
+                    enable_uuid=self._enable_uuid,
                     echo=self._echo,
                     **self._conn_kwargs)
                 # raise exception if pool is closing
@@ -192,13 +228,13 @@ class Pool(asyncio.AbstractServer):
 
     def release(self, conn):
         """Release free connection back to the connection pool.
-
-        This is NOT a coroutine.
         """
+        fut = create_future(self._loop)
+        fut.set_result(None)
         if conn in self._terminated:
             assert conn.closed, conn
             self._terminated.remove(conn)
-            return
+            return fut
         assert conn in self._used, (conn, self._used)
         self._used.remove(conn)
         if not conn.closed:
@@ -208,12 +244,13 @@ class Pool(asyncio.AbstractServer):
                     "Invalid transaction status on released connection: %d",
                     tran_status)
                 conn.close()
-                return
+                return fut
             if self._closing:
                 conn.close()
             else:
                 self._free.append(conn)
-            asyncio.Task(self._wakeup(), loop=self._loop)
+            fut = ensure_future(self._wakeup(), loop=self._loop)
+        return fut
 
     @asyncio.coroutine
     def cursor(self, name=None, cursor_factory=None,
@@ -223,7 +260,7 @@ class Pool(asyncio.AbstractServer):
         cur = yield from conn.cursor(name=name, cursor_factory=cursor_factory,
                                      scrollable=scrollable, withhold=withhold,
                                      timeout=timeout)
-        return _CursorContextManager(self, conn, cur)
+        return _PoolCursorContextManager(self, conn, cur)
 
     def __enter__(self):
         raise RuntimeError(
@@ -248,9 +285,19 @@ class Pool(asyncio.AbstractServer):
         #     finally:
         #         conn.release()
         conn = yield from self.acquire()
-        return _ConnectionContextManager(self, conn)
+        return _PoolConnectionContextManager(self, conn)
 
-    if PY_34:  # pragma: no branch
+    if PY_35:  # pragma: no branch
+        @asyncio.coroutine
+        def __aenter__(self):
+            return self
+
+        @asyncio.coroutine
+        def __aexit__(self, exc_type, exc_val, exc_tb):
+            self.close()
+            yield from self.wait_closed()
+
+    if PY_341:  # pragma: no branch
         def __del__(self):
             try:
                 self._free
@@ -265,70 +312,3 @@ class Pool(asyncio.AbstractServer):
                 warnings.warn(
                     "Unclosed {} connections in {!r}".format(left, self),
                     ResourceWarning)
-
-
-class _ConnectionContextManager:
-    """Context manager.
-
-    This enables the following idiom for acquiring and releasing a
-    connection around a block:
-
-        with (yield from pool) as conn:
-            cur = yield from conn.cursor()
-
-    while failing loudly when accidentally using:
-
-        with pool:
-            <block>
-    """
-
-    __slots__ = ('_pool', '_conn')
-
-    def __init__(self, pool, conn):
-        self._pool = pool
-        self._conn = conn
-
-    def __enter__(self):
-        return self._conn
-
-    def __exit__(self, *args):
-        try:
-            self._pool.release(self._conn)
-        finally:
-            self._pool = None
-            self._conn = None
-
-
-class _CursorContextManager:
-    """Context manager.
-
-    This enables the following idiom for acquiring and releasing a
-    cursor around a block:
-
-        with (yield from pool.cursor()) as cur:
-            yield from cur.execute("SELECT 1")
-
-    while failing loudly when accidentally using:
-
-        with pool:
-            <block>
-    """
-
-    __slots__ = ('_pool', '_conn', '_cur')
-
-    def __init__(self, pool, conn, cur):
-        self._pool = pool
-        self._conn = conn
-        self._cur = cur
-
-    def __enter__(self):
-        return self._cur
-
-    def __exit__(self, *args):
-        try:
-            self._cur._impl.close()
-            self._pool.release(self._conn)
-        finally:
-            self._pool = None
-            self._conn = None
-            self._cur = None
