@@ -1,5 +1,4 @@
 import asyncio
-import weakref
 
 from sqlalchemy.sql import ClauseElement
 from sqlalchemy.sql.dml import UpdateBase
@@ -9,6 +8,7 @@ from . import exc
 from .result import ResultProxy
 from .transaction import (RootTransaction, Transaction,
                           NestedTransaction, TwoPhaseTransaction)
+from ..utils import _SAConnectionContextManager, _TransactionContextManager
 
 
 class SAConnection:
@@ -17,11 +17,9 @@ class SAConnection:
         self._connection = connection
         self._transaction = None
         self._savepoint_seq = 0
-        self._weak_results = weakref.WeakSet()
         self._engine = engine
         self._dialect = engine.dialect
 
-    @asyncio.coroutine
     def execute(self, query, *multiparams, **params):
         """Executes a SQL query with optional parameters.
 
@@ -59,6 +57,11 @@ class SAConnection:
         execution.
 
         """
+        coro = self._execute(query, *multiparams, **params)
+        return _SAConnectionContextManager(coro)
+
+    @asyncio.coroutine
+    def _execute(self, query, *multiparams, **params):
         cursor = yield from self._connection.cursor()
         dp = _distill_params(multiparams, params)
         if len(dp) > 1:
@@ -103,9 +106,7 @@ class SAConnection:
                                     "SQLAlchemy data "
                                     "selection/modification clause")
 
-        ret = ResultProxy(self, cursor, self._dialect)
-        self._weak_results.add(ret)
-        return ret
+        return ResultProxy(self, cursor, self._dialect)
 
     @asyncio.coroutine
     def scalar(self, query, *multiparams, **params):
@@ -126,7 +127,6 @@ class SAConnection:
     def connection(self):
         return self._connection
 
-    @asyncio.coroutine
     def begin(self):
         """Begin a transaction and return a transaction handle.
 
@@ -154,6 +154,11 @@ class SAConnection:
           .begin_twophase - use a two phase/XA transaction
 
         """
+        coro = self._begin()
+        return _TransactionContextManager(coro)
+
+    @asyncio.coroutine
+    def _begin(self):
         if self._transaction is None:
             self._transaction = RootTransaction(self)
             yield from self._begin_impl()
@@ -185,6 +190,7 @@ class SAConnection:
             yield from cur.execute('ROLLBACK')
         finally:
             cur.close()
+            self._transaction = None
 
     @asyncio.coroutine
     def begin_nested(self):
