@@ -4,7 +4,8 @@ import json
 import aiopg
 from .connection import SAConnection
 from .exc import InvalidRequestError
-from aiopg.connection import TIMEOUT
+from ..utils import PY_35, _PoolContextManager, _PoolAcquireContextManager
+from ..connection import TIMEOUT
 
 
 try:
@@ -23,8 +24,7 @@ _dialect.supports_sane_multi_rowcount = True  # psycopg 2.0.9+
 _dialect._has_native_hstore = True
 
 
-@asyncio.coroutine
-def create_engine(dsn=None, *, minsize=10, maxsize=10, loop=None,
+def create_engine(dsn=None, *, minsize=1, maxsize=10, loop=None,
                   dialect=_dialect, timeout=TIMEOUT, **kwargs):
     """A coroutine for Engine creation.
 
@@ -33,6 +33,15 @@ def create_engine(dsn=None, *, minsize=10, maxsize=10, loop=None,
     The pool has *minsize* opened connections to PostgreSQL server.
     """
 
+    coro = _create_engine(dsn=dsn, minsize=minsize, maxsize=maxsize,
+                          loop=loop, dialect=dialect, timeout=timeout,
+                          **kwargs)
+    return _EngineContextManager(coro)
+
+
+@asyncio.coroutine
+def _create_engine(dsn=None, *, minsize=1, maxsize=10, loop=None,
+                   dialect=_dialect, timeout=TIMEOUT, **kwargs):
     if loop is None:
         loop = asyncio.get_event_loop()
     pool = yield from aiopg.create_pool(dsn, minsize=minsize, maxsize=maxsize,
@@ -42,7 +51,7 @@ def create_engine(dsn=None, *, minsize=10, maxsize=10, loop=None,
         real_dsn = conn.dsn
         return Engine(dialect, pool, real_dsn)
     finally:
-        pool.release(conn)
+        yield from pool.release(conn)
 
 
 class Engine:
@@ -99,6 +108,10 @@ class Engine:
     def freesize(self):
         return self._pool.freesize
 
+    @property
+    def closed(self):
+        return self._pool.closed
+
     def close(self):
         """Close engine.
 
@@ -120,9 +133,13 @@ class Engine:
         """Wait for closing all engine's connections."""
         yield from self._pool.wait_closed()
 
-    @asyncio.coroutine
     def acquire(self):
         """Get a connection from pool."""
+        coro = self._acquire()
+        return _EngineAcquireContextManager(coro, self)
+
+    @asyncio.coroutine
+    def _acquire(self):
         raw = yield from self._pool.acquire()
         conn = SAConnection(raw, self)
         return conn
@@ -133,7 +150,8 @@ class Engine:
             raise InvalidRequestError("Cannot release a connection with "
                                       "not finished transaction")
         raw = conn.connection
-        self._pool.release(raw)
+        fut = self._pool.release(raw)
+        return fut
 
     def __enter__(self):
         raise RuntimeError(
@@ -160,6 +178,20 @@ class Engine:
         conn = yield from self.acquire()
         return _ConnectionContextManager(self, conn)
 
+    if PY_35:  # pragma: no branch
+        @asyncio.coroutine
+        def __aenter__(self):
+            return self
+
+        @asyncio.coroutine
+        def __aexit__(self, exc_type, exc_val, exc_tb):
+            self.close()
+            yield from self.wait_closed()
+
+
+_EngineContextManager = _PoolContextManager
+_EngineAcquireContextManager = _PoolAcquireContextManager
+
 
 class _ConnectionContextManager:
     """Context manager.
